@@ -376,6 +376,55 @@ class CaptureStillCommand(Command):
         return self._make_command('Capt', b'')
 
 
+class MediaplayerSelectCommand(Command):
+    """
+    Implementation of the `MPSS` command. This sets the still or clip from the media pool to load into a mediaplayer
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      1    u8     Mask
+    1      1    u8     Mediaplayer index
+    2      1    u8     Source type, 1=still, 2=clip
+    3      1    u8     Still index
+    4      1    u8     Clip index
+    5      3    ?      padding
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, index, still=None, clip=None):
+        """
+        :param index: Mediaplayer index
+        :param still: Still index to load
+        :param clip: Clip index to load
+        """
+        self.index = index
+        self.still = still
+        self.clip = clip
+        if still is not None and clip is not None:
+            raise ValueError("Can only set still= or clip=, not both")
+        if still is None and clip is None:
+            raise ValueError("Either still or clip is required")
+        if clip is None:
+            self.source_type = 1
+        else:
+            self.source_type = 2
+
+    def get_command(self):
+        mask = 1
+        if self.still is not None:
+            mask |= 1 << 1
+        if self.clip is not None:
+            mask |= 1 << 2
+
+        still = self.still if self.still is not None else 0
+        clip = self.clip if self.clip is not None else 0
+
+        data = struct.pack('>BBBBBxxx', mask, self.index, self.source_type, still, clip)
+        return self._make_command('MPSS', data)
+
+
 class DkeyOnairCommand(Command):
     """
     Implementation of the `CDsL` command. This setting the "on-air" state of the downstream keyer on or off
@@ -1246,8 +1295,8 @@ class FairlightStripPropertiesCommand(Command):
 
     """
 
-    def __init__(self, source, channel, delay=None, gain=None, eq_gain=None, dynamics_gain=None, balance=None,
-                 volume=None, state=None):
+    def __init__(self, source, channel, delay=None, gain=None, eq_gain=None, eq_enable=None, dynamics_gain=None,
+                 balance=None, volume=None, state=None):
         """
         :param index: 0-indexed M/E number to control the preview bus of
         """
@@ -1256,6 +1305,7 @@ class FairlightStripPropertiesCommand(Command):
         self.delay = delay
         self.gain = gain
         self.eq_gain = eq_gain
+        self.eq_enable = eq_enable
         self.dynamics_gain = dynamics_gain
         self.balance = balance
         self.volume = volume
@@ -1267,6 +1317,8 @@ class FairlightStripPropertiesCommand(Command):
             mask |= 1 << 0
         if self.gain is not None:
             mask |= 1 << 1
+        if self.eq_enable is not None:
+            mask |= 1 << 3
         if self.eq_gain is not None:
             mask |= 1 << 4
         if self.dynamics_gain is not None:
@@ -1280,6 +1332,7 @@ class FairlightStripPropertiesCommand(Command):
 
         delay = 0 if self.delay is None else self.delay
         gain = 0 if self.gain is None else self.gain
+        eq_enable = False if self.eq_enable is None else self.eq_enable
         eq_gain = 0 if self.eq_gain is None else self.eq_gain
         dynamics_gain = 0 if self.dynamics_gain is None else self.dynamics_gain
         balance = 0 if self.balance is None else self.balance
@@ -1289,8 +1342,9 @@ class FairlightStripPropertiesCommand(Command):
         split = 0xff if self.channel > -1 else 0x01
         self.channel = 0x00 if self.channel == -1 else self.channel
         pad = b'\xff\xff\xff\xff\xff\xff\xff'
-        data = struct.pack('>H H4x6sBb B 3x i 6x h 2x Hh 2x iB 3x', mask, self.source, pad, split, self.channel, delay,
-                           gain, eq_gain,
+        data = struct.pack('>H H4x6sBb B 3x i ? 5x h 2x Hh 2x iB 3x', mask, self.source, pad, split, self.channel,
+                           delay,
+                           gain, eq_enable, eq_gain,
                            dynamics_gain, balance, volume, state)
         return self._make_command('CFSP', data)
 
@@ -1676,6 +1730,198 @@ class KeyPropertiesDveCommand(Command):
         return self._make_command('CKDV', data)
 
 
+class KeyPropertiesAdvancedChromaColorpickerCommand(Command):
+    """
+    Implementation of the `CACC` command. This sets the state of the colorpicker for the advanced chroma keyer
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      1    u8     Mask
+    1      1    u8     M/E index, 0-indexed
+    2      1    u8     Keyer index
+    3      1    bool   Enable cursor
+    4      1    bool   Enable preview
+    5      1    ?      padding
+    6      2    i16    Cursor X [-16000 - 16000]
+    8      2    i16    Cursor Y [-9000 - 9000]
+    10     2    u16    Cursor size [620 - 9925]
+    12     2    u16    Color Y
+    14     2    i16    Color Cb
+    16     2    i16    Color Cr
+    18     2    ?      padding
+    ====== ==== ====== ===========
+
+    === ==========
+    Bit Mask value
+    === ==========
+    0   Enable cursor
+    1   Enable preview
+    2   Cursor X
+    3   Cursor Y
+    4   Cursor size
+    5   Color Y
+    6   Color Cb
+    7   color Cr
+    === ==========
+
+    """
+
+    def __init__(self, index, keyer, cursor=None, preview=None, x=None, y=None, size=None, Y=None, Cb=None, Cr=None):
+        """
+        :param index: 0-indexed M/E number to control the preview bus of
+        """
+        self.index = index
+        self.keyer = keyer
+        self.cursor = cursor
+        self.preview = preview
+        self.x = x
+        self.y = y
+        self.size = size
+        self.Y = Y
+        self.Cb = Cb
+        self.Cr = Cr
+
+    def get_command(self):
+        mask = 0
+        if self.cursor is not None:
+            mask |= 1 << 0
+        if self.preview is not None:
+            mask |= 1 << 1
+        if self.x is not None:
+            mask |= 1 << 2
+        if self.y is not None:
+            mask |= 1 << 3
+        if self.size is not None:
+            mask |= 1 << 4
+        if self.Y is not None:
+            mask |= 1 << 5
+        if self.Cb is not None:
+            mask |= 1 << 6
+        if self.Cr is not None:
+            mask |= 1 << 7
+
+        index = 0 if self.index is None else self.index
+        keyer = 0 if self.keyer is None else self.keyer
+        cursor = False if self.cursor is None else self.cursor
+        preview = False if self.preview is None else self.preview
+        x = 0 if self.x is None else self.x
+        y = 0 if self.y is None else self.y
+        size = 0 if self.size is None else self.size
+        Y = 0 if self.Y is None else self.Y
+        Cb = 0 if self.Cb is None else self.Cb
+        Cr = 0 if self.Cr is None else self.Cr
+
+        data = struct.pack('>BBB? ?x hhH Hhh xx', mask, index, keyer,
+                           cursor, preview, x, y, size, Y, Cb, Cr)
+
+        return self._make_command('CACC', data)
+
+
+class KeyPropertiesAdvancedChromaCommand(Command):
+    """
+    Implementation of the `CACK` command. This sets the state for the advanced chroma keyer
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      2    u16    Mask
+    2      1    u8     M/E index, 0-indexed
+    3      1    u8     Keyer index
+    4      2    u16    Foreground [0 - 1000]
+    6      2    u16    Background [0 - 1000]
+    8      2    u16    Key edge [0 - 1000]
+    10     2    u16    Spill suppress [0 - 1000]
+    12     2    u16    Flare suppress [0 - 1000]
+    14     2    i16    Brightness [-1000 - 1000]
+    16     2    i16    Contrast [-1000 - 1000]
+    18     2    u16    Saturation [0 - 2000]
+    20     2    i16    Red [-1000 - 1000]
+    22     2    i16    Green [-1000 - 1000]
+    24     2    i16    Blue [-1000 - 1000]
+    26     2    ?      padding
+    ====== ==== ====== ===========
+
+    === ==========
+    Bit Mask value
+    === ==========
+    0   Foreground
+    1   Background
+    2   Key edge
+    3   Spill suppress
+    4   Flare suppress
+    5   Brightness
+    6   Contrast
+    7   Saturation
+    8   Red
+    9   Green
+    10  Blue
+    === ==========
+
+    """
+
+    def __init__(self, index, keyer, foreground=None, background=None, key_edge=None, spill=None, flare=None,
+                 brightness=None, contrast=None, saturation=None, red=None, green=None, blue=None):
+        """
+        :param index: 0-indexed M/E number to control the preview bus of
+        """
+        self.index = index
+        self.keyer = keyer
+        self.foreground = foreground
+        self.background = background
+        self.key_edge = key_edge
+        self.spill = spill
+        self.flare = flare
+        self.brightness = brightness
+        self.contrast = contrast
+        self.saturation = saturation
+        self.red = red
+        self.green = green
+        self.blue = blue
+
+    def get_command(self):
+        mask = 0
+        if self.foreground is not None:
+            mask |= 1 << 0
+        if self.background is not None:
+            mask |= 1 << 1
+        if self.key_edge is not None:
+            mask |= 1 << 2
+        if self.spill is not None:
+            mask |= 1 << 3
+        if self.flare is not None:
+            mask |= 1 << 4
+        if self.brightness is not None:
+            mask |= 1 << 5
+        if self.contrast is not None:
+            mask |= 1 << 6
+        if self.saturation is not None:
+            mask |= 1 << 7
+        if self.red is not None:
+            mask |= 1 << 8
+        if self.green is not None:
+            mask |= 1 << 9
+        if self.blue is not None:
+            mask |= 1 << 10
+
+        foreground = 0 if self.foreground is None else self.foreground
+        background = 0 if self.background is None else self.background
+        key_edge = 0 if self.key_edge is None else self.key_edge
+        spill = 0 if self.spill is None else self.spill
+        flare = 0 if self.flare is None else self.flare
+        brightness = 0 if self.brightness is None else self.brightness
+        contrast = 0 if self.contrast is None else self.contrast
+        saturation = 0 if self.saturation is None else self.saturation
+        red = 0 if self.red is None else self.red
+        green = 0 if self.green is None else self.green
+        blue = 0 if self.blue is None else self.blue
+
+        data = struct.pack('>HBB HHH HH hhHhhh xx', mask, self.index, self.keyer, foreground, background, key_edge, spill,
+                           flare, brightness, contrast, saturation, red, green, blue)
+
+        return self._make_command('CACK', data)
+
+
 class KeyPropertiesLumaCommand(Command):
     """
     Implementation of the `CKLm` command. This sets the key source for an upstream keyer.
@@ -1896,10 +2142,20 @@ class StreamingServiceSetCommand(Command):
     1      64   str    Service name
     65     512  str    Url
     577    512  str    Streaming key
+    1089   3    ?      padding bytes
     1092   4    u32    Minimum bitrate (Bps)
     1096   4    u32    Maximum bitrate (Bps)
     ====== ==== ====== ===========
 
+    === ==========
+    Bit Mask value
+    === ==========
+    0   Service name
+    1   Url
+    2   Key
+    3   Min/Max bitrate
+    === ==========
+
     """
 
     def __init__(self, name=None, url=None, key=None, bitrate_min=None, bitrate_max=None):
@@ -2050,9 +2306,8 @@ class LockCommand(Command):
 
     def __init__(self, store, state):
         """
-        :param index: 0-indexed DSK number to trigger
-        :param keyer: 0-indexed keyer number
-        :param source: Source index for the keyer fill
+        :param store: Store number to get the lock for
+        :param state: True to request the lock, False to release it
         """
         self.store = store
         self.state = state
@@ -2062,6 +2317,33 @@ class LockCommand(Command):
         return self._make_command('LOCK', data)
 
 
+class PartialLockCommand(Command):
+    """
+    Implementation of the `PLCK` command. This requests a new lock, used for data transfers.
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      1    u16    Store index
+    2      2    u16    Slot
+    4      4    ?      unknown
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, store, slot):
+        """
+        :param store: Store number to request a lock for
+        :param slot: Slot number inside the store to get the partial lock for
+        """
+        self.store = store
+        self.slot = slot
+
+    def get_command(self):
+        data = struct.pack('>HH BBxx', self.store, self.slot, 0xff, 0x01)
+        return self._make_command('PLCK', data)
+
+
 class TransferDownloadRequestCommand(Command):
     """
     Implementation of the `FTSU` command. This requests download from the switcher to the client.
@@ -2100,6 +2382,119 @@ class TransferDownloadRequestCommand(Command):
         return self._make_command('FTSU', data)
 
 
+class TransferUploadRequestCommand(Command):
+    """
+    Implementation of the `FTSD` command. This requests an upload from the client to the switcher.
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      2    u16    Tranfer id
+    2      2    u16    Store
+    4      2    ?      unknown
+    6      2    u16    Slot
+    8      2    u32    Data length
+    12     2    u16    Mode bitfield
+    14     2    ?      unknown
+    ====== ==== ====== ===========
+
+    === ==========
+    bit Mode
+    === ==========
+    0   Unknown
+    1   Write RLE
+    2   Clear
+    256 Write uncompressed
+    512 Pre-erase
+    === ==========
+
+    """
+
+    MODE_WRITE_RLE = 1
+    MODE_WRITE = 256
+    MODE_ERASE = 512
+
+    def __init__(self, transfer, store, slot, length, mode):
+        """
+        :param transfer: Unique transfer number
+        :param store: Store index
+        :param slot: Slot index
+        """
+        self.transfer = transfer
+        self.store = store
+        self.slot = slot
+        self.length = length
+        self.mode = mode
+
+    def get_command(self):
+        data = struct.pack('>HHxxHIHxx', self.transfer, self.store, self.slot, self.length, self.mode)
+        return self._make_command('FTSD', data)
+
+
+class TransferDataCommand(Command):
+    """
+    Implementation of the `FTDa` command. This is a chunk of data being sent from the client to the hardware.
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      2    u16    Tranfer id
+    2      2    u16    Size
+    4      ?    u8[]   Chunk of data
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, transfer, data):
+        """
+        :param transfer: Unique transfer number
+        :param data: Chunk of data to send
+        """
+        self.transfer = transfer
+        self.data = data
+
+    def get_command(self):
+        data = struct.pack('>HH', self.transfer, len(self.data))
+        return self._make_command('FTDa', data + self.data)
+
+
+class TransferFileDataCommand(Command):
+    """
+    Implementation of the `FTFD` command. This is the file metadata for the uploaded file and should be sent to the
+    hardware after all the FTDa commands have been sent. It contains the name and MD5 hash of the uploaded data.
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      2    u16    Tranfer id
+    2      64   str    Name
+    66     128  str    Description
+    194    16   u8[]   MD5 hash
+    210    2    ?      unknown
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, transfer, hash, name=None, description=None):
+        """
+        :param transfer: Unique transfer number
+        :param name: Filename
+        :param description: File description
+        :param hash: Data MD5 hash
+        """
+        self.transfer = transfer
+        self.name = name
+        self.description = description
+        self.hash = hash
+
+    def get_command(self):
+        name = self.name.encode() if self.name is not None else b''
+        description = self.description.encode() if self.description is not None else b''
+
+        data = struct.pack('>H 64s 128s 16s 2x', self.transfer, name, description, self.hash)
+        return self._make_command('FTFD', data)
+
+
 class TransferAckCommand(Command):
     """
     Implementation of the `FTUA` command. This is an acknowledgement for FTDa packets.
@@ -2252,3 +2647,140 @@ class CameraControlCommand(Command):
             packed_data += b'\0' * (8 - len(packed_data))
             data += packed_data
         return self._make_command('CCmd', data)
+
+
+class VideoModeCommand(Command):
+    """
+    Implementation of the `CVdM` command. This sets the main video mode for the hardware
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      1    u8     M/E index
+    1      1    bool   Preview enabled
+    2      2    ?      unknown
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, mode):
+        """
+        :param mode: The new video mode ID
+        """
+        self.mode = mode
+
+    def get_command(self):
+        data = struct.pack('>B3x', self.mode)
+        return self._make_command('CVdM', data)
+
+
+class AutoInputVideoModeCommand(Command):
+    """
+    Implementation of the `AiVM` command. This enables or disables automatic video mode detection based on the video
+    signal.
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      1    bool   Enable
+    1      3    ?      unknown
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, enable):
+        """
+        :param enable: Set auto mode enabled or disabled
+        """
+        self.enable = enable
+
+    def get_command(self):
+        data = struct.pack('>?3x', self.enable)
+        return self._make_command('AiVM', data)
+
+
+class InputPropertiesCommand(Command):
+    """
+    Implementation of the `CInL` command. This sets labels and routing for video inputs
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      1    u8     Mask
+    1      1    ?      unknown
+    2      2    u16    Source index
+    4      20   str[]  Label
+    24     4    str[]  Short label
+    28     2    u16    Port type
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, source_index, label=None, short_label=None, port_type=None):
+        """
+        :param source_index: Input index to change the properties for
+        :param label: Long display label for the input
+        :param short_label: Short display label for the button of the input, 4 letter code
+        :param port_type: Port type enum
+        """
+        self.source_index = source_index
+        self.label = label
+        self.short_label = short_label
+        self.port_type = port_type
+
+    def get_command(self):
+        mask = 0
+        if self.label is not None:
+            mask |= 1 << 0
+        if self.short_label is not None:
+            mask |= 1 << 1
+        if self.port_type is not None:
+            mask |= 1 << 2
+
+        label = self.label.encode() if self.label is not None else b''
+        short = self.short_label.encode() if self.short_label is not None else b''
+        port_type = self.port_type if self.port_type is not None else 0
+
+        data = struct.pack('>Bx H 20s 4s Hxx', mask, self.source_index, label, short, port_type)
+        return self._make_command('CInL', data)
+
+
+class TimeRequestCommand(Command):
+    """
+    Request the system time of the hardware, this is also used as a NOP command
+    This command has no arguments
+    """
+
+    def get_command(self):
+        return self._make_command('TiRq', b'')
+
+
+class TransferCompleteCommand(Command):
+    """
+    Implementation of the `*XFC` command. This is an command that's part of OpenSwitcher for the TCP protocol and not part
+    of the actual ATEM protocol.
+
+    ====== ==== ====== ===========
+    Offset Size Type   Description
+    ====== ==== ====== ===========
+    0      2    u16    Store
+    2      2    u16    Slot
+    4      1    bool   Is upload
+    5      3    x      padding
+    ====== ==== ====== ===========
+
+    """
+
+    def __init__(self, store, slot, upload):
+        """
+        :param store: Transfer store index
+        :param slot: Transfer slot index
+        :param upload: True if an upload was completed, False if a download was completed
+        """
+        self.store = store
+        self.slot = slot
+        self.upload = upload
+
+    def get_command(self):
+        data = struct.pack('>HH ?xxx', self.store, self.slot, self.upload)
+        return self._make_command('*XFC', data)
